'''Linters for various languages.'''
# pylint: disable=too-many-lines

from __future__ import print_function

from html.parser import HTMLParser
import bisect
import dataclasses
import importlib.util
import json
import logging
import os
import os.path
import re
import shlex
import socket
import subprocess
import sys
import tempfile
import threading
from typing import (Any, Callable, Dict, List, Mapping, NamedTuple, Optional,
                    Pattern, Text, Sequence, Tuple)

from . import git_tools


def _unicode_whitespace() -> Text:
    '''Returns a string constant with all whitespace characters, sans \n.'''
    return repr(''.join(
        re.findall(
            r'\s',
            ''.join(chr(c) for c in range(sys.maxunicode + 1)
                    if c != ord('\n')))))[1:-1]


def _find_pip_tool(name: Text) -> Text:
    '''Tries to find a pip tool in a few default locations.'''
    for prefix in ['/usr/bin', '/usr/local/bin']:
        toolpath = os.path.join(prefix, name)
        if os.path.exists(toolpath):
            return toolpath
    return os.path.join(os.environ['HOME'], '.local/bin', name)


def _which(program: Text) -> Text:
    '''Looks for |program| in $PATH. Similar to UNIX's `which` command.'''
    for path in ['./node_modules/.bin'] + os.environ['PATH'].split(os.pathsep):
        exe_file = os.path.abspath(os.path.join(path.strip('"'), program))
        if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
            return exe_file
    raise Exception(f'`{program}` not found')


_TIDY_PATH = os.path.join(git_tools.HOOK_TOOLS_ROOT, 'data/tidy')
_PRETTIER_RE = re.compile(r'^(?:[^\s:]+\s*)?[^\s:]+: (.*) \((\d+):(\d+)\)\s*$')
_DIAGNOSTIC_RE = re.compile(r'^([^:\s]+):(\d+):(\d+): (.*)$')

Options = Mapping[str, Any]
ContentsCallback = Callable[[str], bytes]


class SingleResult(NamedTuple):
    '''The result of Linter.run_one().'''
    contents: bytes
    violations: Sequence[str] = ()


class MultipleResults(NamedTuple):
    '''The result of Linter.run_all().'''
    new_contents: Mapping[str, bytes]
    original_contents: Mapping[str, bytes]
    violations: Sequence[str] = ()


class ProblematicTerm(NamedTuple):
    '''Represents a list of problematic terms/messages associated with them.'''
    regexps: Sequence[Pattern[bytes]]
    message: str
    allowlist: Optional[Sequence[Pattern[str]]]
    denylist: Optional[Sequence[Pattern[str]]]


@dataclasses.dataclass
class Diagnostic:
    '''A diagnostic message.'''
    message: str
    filename: str
    line: Optional[str] = None
    lineno: Optional[int] = None
    col: Optional[int] = None
    col_end: Optional[int] = None
    level: str = 'error'


class LinterException(Exception):
    '''A fatal exception during linting.'''

    def __init__(
            self,
            message: Text,
            fixable: bool = True,
            diagnostics: Sequence[Diagnostic] = ()
    ) -> None:
        super().__init__(message)
        self.__message = message
        self.__fixable = fixable
        self.__diagnostics = diagnostics

    @property
    def message(self) -> Text:
        '''A message that can be presented to the user.'''
        return self.__message

    @property
    def fixable(self) -> bool:
        '''Whether this exception supports being fixed.'''
        return self.__fixable

    @property
    def diagnostics(self) -> Sequence[Diagnostic]:
        '''Any diagnostics that this LinterException could have.'''
        return self.__diagnostics


def _process_diagnostics_output(
        filename: str,
        lines: Sequence[str],
        output: str,
        toolname: Optional[str] = None) -> Sequence[Diagnostic]:
    '''Process the output of a tool in standard format.'''
    diagnostics: List[Diagnostic] = []
    for line in output.split('\n'):
        match = _DIAGNOSTIC_RE.match(line.strip())
        if not match:
            continue
        if os.path.basename(match.group(1)) != os.path.basename(filename):
            # Some diagnostics will refer to other code locations, maybe in
            # other files.
            continue
        highlighted_line = ''
        lineno = int(match.group(2))
        if len(lines) >= lineno:
            highlighted_line = lines[lineno - 1].rstrip()
        diagnostics.append(
            Diagnostic(
                (f'[{toolname}] {match.group(4)}'
                 if toolname else match.group(4)),
                filename=filename,
                lineno=lineno,
                line=highlighted_line,
                col=int(match.group(3)) or None,
            ))
    return diagnostics


def _custom_command(command: Text, filename: Text,
                    original_filename: Text) -> List[Text]:
    '''A custom command.'''

    return [
        '/bin/bash', '-c',
        f'{command} {shlex.quote(filename)} {shlex.quote(original_filename)}',
    ]


def _lint_javascript(filename: Text,
                     contents: bytes,
                     extra_commands: Optional[Sequence[Text]] = None) -> bytes:
    '''Runs prettier on |contents|.'''

    lines = contents.decode('utf-8').split('\n')
    with tempfile.NamedTemporaryFile(suffix='.js') as js_out:
        # Keep the shebang unmodified.
        header = b''
        if contents.startswith(b'#!'):
            header, contents = contents.split(b'\n', 1)
            header += b'\n'

        js_out.write(contents)
        js_out.flush()

        commands = [
            [
                _which('prettier'), '--single-quote', '--trailing-comma=all',
                '--no-config', '--write', js_out.name
            ],
        ] + [
            _custom_command(command, js_out.name, filename)
            for command in (extra_commands or [])
        ]

        for args in commands:
            try:
                logging.debug('lint_javascript: Running %s', args)
                subprocess.run(args,
                               check=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True)
            except subprocess.CalledProcessError as cpe:
                diagnostics: List[Diagnostic] = []
                for line in cpe.output.strip().split('\n'):
                    match = _PRETTIER_RE.match(line)
                    if not match:
                        continue
                    diagnostics.append(
                        Diagnostic(
                            f'{match.group(1)}',
                            filename=filename,
                            lineno=int(match.group(2)),
                            line=lines[int(match.group(2)) - 1].rstrip(),
                            col=int(match.group(3)) or None,
                        ))
                raise LinterException('JavaScript lint errors',
                                      fixable=False,
                                      diagnostics=diagnostics) from cpe

        with open(js_out.name, 'rb') as js_in:
            return header + js_in.read()


def _lint_html(contents: bytes, strict: bool) -> bytes:
    '''Runs tidy on |contents|.'''

    args = [
        _TIDY_PATH, '-q', '-config',
        os.path.join(git_tools.HOOK_TOOLS_ROOT, 'data/tidy.txt')
    ]
    logging.debug('lint_html: Running %s', args)
    result = subprocess.run(args,
                            input=contents,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=False,
                            cwd=git_tools.HOOK_TOOLS_ROOT)

    if result.returncode == 0:
        return result.stdout
    if result.returncode == 1 and not strict:
        # |result.returncode| == 1 means that there were warnings.
        return result.stdout

    raise LinterException(result.stderr.decode('utf-8', errors='replace'))


def _lint_prettier(contents: bytes, filename: Text) -> bytes:
    '''Runs prettier on |contents| .'''
    args = [
        _which('prettier'), '--single-quote', '--trailing-comma=all',
        '--no-config',
        f'--stdin-filepath={filename}'
    ]
    logging.debug('lint_prettier: Running %s', args)
    result = subprocess.run(args,
                            input=contents,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            check=False,
                            cwd=git_tools.HOOK_TOOLS_ROOT)

    if result.returncode != 0:
        lines = contents.decode('utf-8').split('\n')
        diagnostics: List[Diagnostic] = []
        for line in result.stderr.decode('utf-8',
                                         errors='replace').strip().split('\n'):
            match = _PRETTIER_RE.match(line)
            if not match:
                continue
            diagnostics.append(
                Diagnostic(
                    f'{match.group(1)}',
                    filename=filename,
                    lineno=int(match.group(2)),
                    line=lines[int(match.group(2)) - 1].rstrip(),
                    col=int(match.group(3)) or None,
                ))
        raise LinterException('lint errors',
                              fixable=False,
                              diagnostics=diagnostics)

    return result.stdout


def _lint_stylelint(contents: bytes, filename: Text) -> bytes:
    '''Runs stylelint on |contents| .'''
    args = [
        _which('stylelint'),
        '--stdin',
        f'--stdin-filename={filename}',
        '--fix',
    ]
    logging.debug('lint_style: Running %s', args)
    result = subprocess.run(args,
                            input=contents,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            check=False)

    if result.returncode == 0:
        # Everything is correct!
        return contents

    if result.returncode != 2:
        # stylelint returns 1 if something went catastrophically wrong, or 78
        # if there was a problem with the configuration. This means that the
        # only return code that we expect right now is 2, which is returned if
        # there were lint errors.
        raise LinterException(result.stdout.decode('utf-8'))

    if result.stdout != contents:
        # If stylelint was able to automatically fix anything, we go with
        # that.
        return result.stdout

    args = [
        _which('stylelint'),
        '--stdin',
        f'--stdin-filename={filename}',
        '--formatter=unix',
    ]
    logging.debug('lint_style: Running %s', args)
    result = subprocess.run(args,
                            input=contents,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            check=False)

    if result.returncode != 0:
        raise LinterException(
            'lint errors',
            fixable=False,
            diagnostics=_process_diagnostics_output(
                filename,
                lines=contents.decode('utf-8').split('\n'),
                output=result.stdout.decode('utf-8').strip()))

    return contents


class Linter:
    '''An abstract Linter.'''
    # pylint: disable=R0903

    def __init__(self) -> None:
        pass

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        '''Runs the linter against |contents|.'''
        del filename  # unused
        return SingleResult(contents)

    def run_all(self, filenames: Sequence[str],
                contents_callback: ContentsCallback) -> MultipleResults:
        '''Runs the linter against a subset of files.'''
        del filenames, contents_callback  # unused
        return MultipleResults(new_contents={}, original_contents={})

    @property
    def name(self) -> Text:
        '''Gets the name of the linter.'''
        return 'linter'


class JavaScriptLinter(Linter):
    '''Runs the Google Closure Compiler linter+prettier against |files|.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        try:
            return SingleResult(
                _lint_javascript(filename, contents,
                                 self.__options.get('extra_js_linters')),
                ['javascript'])
        except subprocess.CalledProcessError as cpe:
            raise LinterException(
                str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                    encoding='utf-8')) from cpe

    @property
    def name(self) -> Text:
        return 'javascript'


class TypeScriptLinter(Linter):
    '''Runs prettier against |files|.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        try:
            return SingleResult(_lint_prettier(contents, filename),
                                ['typescript'])
        except subprocess.CalledProcessError as cpe:
            raise LinterException(
                str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                    encoding='utf-8')) from cpe

    @property
    def name(self) -> Text:
        return 'typescript'


class KarelLinter(Linter):
    '''Runs prettier against |files|.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        try:
            return SingleResult(_lint_prettier(contents, filename),
                                ['karel'])
        except subprocess.CalledProcessError as cpe:
            raise LinterException(
                str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                    encoding='utf-8')) from cpe

    @property
    def name(self) -> Text:
        return 'karel'


class MarkdownLinter(Linter):
    '''Runs prettier against |files|.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        try:
            return SingleResult(_lint_prettier(contents, filename),
                                ['markdown'])
        except subprocess.CalledProcessError as cpe:
            raise LinterException(
                str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                    encoding='utf-8')) from cpe

    @property
    def name(self) -> Text:
        return 'markdown'


class JsonLinter(Linter):
    '''Runs prettier against |files|.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        try:
            return SingleResult(_lint_prettier(contents, filename),
                                ['json'])
        except subprocess.CalledProcessError as cpe:
            raise LinterException(
                str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                    encoding='utf-8')) from cpe

    @property
    def name(self) -> Text:
        return 'json'


class WhitespaceLinter(Linter):
    '''Removes annoying superfluous whitespace.'''
    # pylint: disable=R0903

    _VALIDATIONS = [
        ('Windows-style EOF', re.compile(br'\r\n?'), br'\n'),
        ('trailing whitespace', re.compile(br'[ \t\r\f\v]+\n'), br'\n'),
        ('consecutive empty lines', re.compile(br'\n\n\n+'), br'\n\n'),
        ('empty lines after an opening brace', re.compile(br'{\n\n+'),
         br'{\n'),
        ('empty lines before a closing brace', re.compile(br'\n+\n(\s*})'),
         br'\n\1'),
    ]

    _UNICODE_VALIDATIONS = [
        ('Windows-style EOF', re.compile(r'\r\n?'), r'\n'),
        ('trailing whitespace',
         re.compile(r'[' + _unicode_whitespace() + r'\u200b\u200c]+\n'),
         r'\n'),
        ('consecutive empty lines', re.compile(r'\n\n\n+'), r'\n\n'),
        ('empty lines after an opening brace', re.compile(r'{\n\n+'), r'{\n'),
        ('empty lines before a closing brace', re.compile(r'\n+\n(\s*})'),
         r'\n\1'),
    ]

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        '''Runs all validations against |files|.

        A validation consists of performing regex substitution against the
        contents of each file in |files|.  Validation fails if the resulting
        content is not identical to the original.  The contents of the files
        will be presented as a single string, allowing for multi-line matches.
        '''
        violations = []

        # Run all validations sequentially, so all violations can be fixed
        # together.
        try:
            unicode_contents = contents.decode('utf-8')
            for (error_string, unicode_search,
                 unicode_replace) in WhitespaceLinter._UNICODE_VALIDATIONS:
                unicode_replaced = unicode_search.sub(unicode_replace,
                                                      unicode_contents)
                if unicode_replaced != unicode_contents:
                    violations.append(error_string)
                    unicode_contents = unicode_replaced
            contents = unicode_contents.encode('utf-8')
        except UnicodeDecodeError:
            for error_string, search, replace in WhitespaceLinter._VALIDATIONS:
                replaced = search.sub(replace, contents)
                if replaced != contents:
                    violations.append(error_string)
                    contents = replaced

        return SingleResult(contents, violations)

    @property
    def name(self) -> Text:
        return 'whitespace'


class ProblematicTermsLinter(Linter):
    '''Warns if there are problematic terms in the codebase.'''
    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__terms: List[ProblematicTerm] = []
        for term in (options or {}).get('terms', []):
            if not isinstance(term, dict):
                raise ValueError(f'{term} is not a dictionary')
            if 'regexps' not in term:
                raise ValueError(f'{term} does not contain a key "regexps"')
            if not isinstance(term['regexps'], list):
                raise ValueError(f'{term} is not a list of string')
            if not all(isinstance(regexp, str) for regexp in term['regexps']):
                raise ValueError(f'{term} is not a list of string')
            allowlist: Optional[Sequence[Pattern[str]]] = None
            denylist: Optional[Sequence[Pattern[str]]] = None
            if 'allowlist' in term:
                allowlist = [
                    re.compile(pattern) for pattern in term['allowlist']
                ]
            if 'denylist' in term:
                denylist = [
                    re.compile(pattern) for pattern in term['denylist']
                ]
            self.__terms.append(
                ProblematicTerm(
                    regexps=[
                        re.compile(regexp.encode('utf-8'), re.MULTILINE)
                        for regexp in term['regexps']
                    ],
                    message=term.get('message', 'is a problematic term'),
                    allowlist=allowlist,
                    denylist=denylist,
                ))

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        '''Tries to identify problematic term in the provided file.'''
        diagnostics = []
        lines = contents.split(b'\n')
        start_indices: List[int] = [0]

        for line in lines:
            start_indices.append(start_indices[-1] + len(line) + 1)

        for term in self.__terms:
            if term.denylist is not None and any(
                    r.match(filename) for r in term.denylist):
                continue
            if term.allowlist is not None and not any(
                    r.match(filename) for r in term.allowlist):
                continue
            for regexp in term.regexps:
                for match in regexp.finditer(contents):
                    found_term = match.group()
                    lineno = bisect.bisect_left(start_indices, match.start())
                    col = match.start() - start_indices[lineno] + 1
                    diagnostics.append(
                        Diagnostic(
                            f'{found_term.decode("utf-8")}: {term.message}',
                            filename=filename,
                            lineno=lineno + 1,
                            col=col,
                            col_end=col + len(found_term)))

        if diagnostics:
            raise LinterException('Found problematic terms',
                                  diagnostics=diagnostics,
                                  fixable=False)

        return SingleResult(contents, [])

    @property
    def name(self) -> Text:
        return 'problematic_terms'


class VueHTMLParser(HTMLParser):
    '''A parser that can understand .vue template files.'''

    # pylint: disable=R0903

    def __init__(self) -> None:
        super().__init__()
        self._stack: List[Tuple[Text, List[Tuple[str, Optional[str]]], Text,
                                Tuple[int, int]]] = []
        self._tags: List[Tuple[Text, List[Tuple[str, Optional[str]]], str,
                               Tuple[int, int], Tuple[int, int]]] = []
        self._id_linter_enabled = True
        self.diagnostics: List[Diagnostic] = []
        self._lines: List[str] = []
        self._filename: str = ''

    def error(self, message: Text) -> None:
        '''Raises an error given the message.'''

        raise LinterException(message)

    def parse(
            self, contents: Text, filename: str
    ) -> Sequence[Tuple[Text, List[Tuple[str, Optional[str]]], Text, Text]]:
        '''Parses |contents| and returns the .vue-specific sections.'''

        self._stack = []
        self._tags = []
        self._filename = filename
        self._lines = contents.split('\n')
        self.feed(contents)

        sections = []
        for tag, attrs, starttag, start, end in self._tags:
            line_range = []
            if len(self._lines[start[0]]) > len(starttag) + start[1]:
                line_range.append(self._lines[start[0]][len(starttag)
                                                        + start[1]:])
            line_range += self._lines[start[0] + 1:end[0]]
            if end[1] > 0:
                line_range.append(self._lines[end[0]][:end[1]])
            sections.append((tag, attrs, starttag, '\n'.join(line_range)))
        return sections

    def handle_starttag(self, tag: Text,
                        attrs: List[Tuple[Text, Optional[Text]]]) -> None:
        line, col = self.getpos()
        self._stack.append((tag, attrs, str(self.get_starttag_text()),
                            (line - 1, col)))
        if not self._id_linter_enabled:
            return
        for name, _ in attrs:
            if name == 'id':
                self.diagnostics.append(
                    Diagnostic(
                        'Use of "id" attribute in .vue files is discouraged.',
                        filename=self._filename,
                        line=self._lines[line - 1],
                        lineno=line,
                        col=col + 1))

    def handle_endtag(self, tag: Text) -> None:
        while self._stack and self._stack[-1][0] != tag:
            self._stack.pop()
        if not self._stack or self._stack[-1][0] != tag:
            line, col = self.getpos()
            self.diagnostics.append(
                Diagnostic('Unclosed tag',
                           filename=self._filename,
                           line=self._lines[line - 1],
                           lineno=line,
                           col=col + 1))
            return
        _, attrs, starttag, begin = self._stack.pop()
        if not self._stack:
            line, col = self.getpos()
            self._tags.append((tag, attrs, starttag, begin, (line - 1, col)))

    def handle_comment(self, data: Text) -> None:
        if data.find('id-lint ') < 0:
            return
        self._id_linter_enabled = data.strip().split()[1] == 'on'


class VueLinter(Linter):
    '''A linter for .vue files.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        parser = VueHTMLParser()
        try:
            sections = parser.parse(contents.decode('utf-8'),
                                    filename=filename)
        except AssertionError as assertion:
            raise LinterException(str(assertion),
                                  diagnostics=parser.diagnostics,
                                  fixable=False) from assertion
        if parser.diagnostics:
            raise LinterException('Found Vue errors',
                                  diagnostics=parser.diagnostics,
                                  fixable=False)

        new_sections = []
        for tag, _, starttag, section_contents in sorted(
                sections,
                key=lambda x: ['template', 'script', 'style'].index(x[0])):
            try:
                new_sections.append(
                    f'{starttag}\n{section_contents.rstrip()}\n</{tag}>')
            except subprocess.CalledProcessError as cpe:
                raise LinterException(
                    str(b'\n'.join(cpe.output.split(b'\n')[1:]),
                        encoding='utf-8'),
                    fixable=False) from cpe

        if len(new_sections) != len(sections):
            raise LinterException(
                (f'Mismatched sections: expecting {len(sections)}, '
                 f'got {len(new_sections)}'),
                fixable=False)

        contents = '\n\n'.join(new_sections).encode('utf-8') + b'\n'
        if self.__options.get('stylelint', False):
            contents = _lint_stylelint(contents, filename)

        # Finally, run prettier on the whole thing.
        return SingleResult(_lint_prettier(contents, filename), ['vue'])

    @property
    def name(self) -> Text:
        return 'vue'


class StyleLinter(Linter):
    '''A linter for .css/.scss files.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        contents = _lint_stylelint(contents, filename)

        # Finally, run prettier on the whole thing.
        return SingleResult(_lint_prettier(contents, filename), ['style'])

    @property
    def name(self) -> Text:
        return 'style'


class HTMLLinter(Linter):
    '''Runs HTML Tidy + prettier.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        contents = _lint_html(
            contents, strict=bool(self.__options.get('strict', False)))

        # Finally, run prettier on the whole thing.
        return SingleResult(_lint_prettier(contents, filename), ['html'])

    @property
    def name(self) -> Text:
        return 'html'


class PHPLinter(Linter):
    '''Runs the PHP Code Beautifier.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}
        standard = self.__options.get(
            'standard',
            os.path.join(git_tools.HOOK_TOOLS_ROOT,
                         'data/phpcbf/Standards/OmegaUp/ruleset.xml'))
        self.__common_args = ['--encoding=utf-8', f'--standard={standard}']

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        args = ([_which('phpcbf')] + self.__common_args
                + [f'--stdin-path={filename}'])
        logging.debug('lint_php: Running %s', shlex.join(args))
        result = subprocess.run(args,
                                input=contents,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                check=False)

        if result.returncode != 0:
            logging.debug('lint_php: Return code %d, stderr = %s',
                          result.returncode, result.stderr)
            if not result.stdout:
                # phpcbf returns 1 if there was no change to the file. If there
                # was an actual error, there won't be anything in stdout.
                raise LinterException(result.stderr.decode('utf-8'))

        new_contents = result.stdout

        if new_contents != contents:
            # If phpcbf was able to fix anything, let's go with that instead of
            # running phpcs. Otherwise, phpcs will return non-zero and the
            # suggestions won't be used.
            return SingleResult(new_contents, ['php'])

        # Even if phpcbf didn't find anything, phpcs might.
        args = ([_which('phpcs'), '-s', '-q'] + self.__common_args
                + [f'--stdin-path={filename}', '--report=emacs'])
        logging.debug('lint_php: Running %s', shlex.join(args))
        result = subprocess.run(args,
                                input=contents,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                check=False)

        if result.returncode != 0:
            logging.debug('lint_php: Return code %d, stdout = %s, stderr = %s',
                          result.returncode, result.stdout, result.stderr)
            raise LinterException(
                'phpcs lint errors',
                fixable=False,
                diagnostics=_process_diagnostics_output(
                    filename,
                    lines=contents.decode('utf-8').split('\n'),
                    output=result.stdout.decode('utf-8').strip()))

        return SingleResult(new_contents, ['php'])

    @property
    def name(self) -> Text:
        return 'php'


class PythonLinter(Linter):
    '''Runs pycodestyle, pylint, and Mypy.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        diagnostics: List[Diagnostic] = []
        lines = contents.decode('utf-8').split('\n')
        with tempfile.TemporaryDirectory(prefix='python_linter_') as tmpdir:
            tmp_path = os.path.join(tmpdir, os.path.basename(filename))
            with open(tmp_path, 'wb') as pyfile:
                pyfile.write(contents)

            python3 = _which('python3')

            args = [
                python3, '-m', 'pycodestyle',
                '--format=%(path)s:%(row)d:%(col)d: %(code)s %(text)s',
                tmp_path
            ]
            for configname in ('pycodestyle_config', 'pep8_config'):
                if configname not in self.__options:
                    continue
                args.append(f'--config={self.__options[configname]}')
                break
            try:
                logging.debug('lint_python: Running %s', args)
                subprocess.run(args,
                               check=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True)
            except subprocess.CalledProcessError as cpe:
                diagnostics.extend(
                    _process_diagnostics_output(filename,
                                                lines,
                                                cpe.output.strip(),
                                                toolname='pycodestyle'))

            # We need to disable import-error/no-name-in-module since the file
            # won't be checked in the repository, but in a temporary directory.
            args = [
                python3,
                '-m',
                'pylint',
                '--output-format=text',
                ('--msg-template={abspath}:{line}:{column}: '
                 '{msg_id}({symbol}) {msg}'),
                '--reports=no',
                '--disable=import-error',
                '--disable=no-name-in-module',
                tmp_path,
            ]
            if 'pylint_config' in self.__options:
                args.append(f'--rcfile={self.__options["pylint_config"]}')
            try:
                logging.debug('lint_python: Running %s', args)
                subprocess.run(args,
                               check=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True)
            except subprocess.CalledProcessError as cpe:
                diagnostics.extend(
                    _process_diagnostics_output(filename,
                                                lines,
                                                cpe.output.strip(),
                                                toolname='pylint'))

            if self.__options.get('mypy', False):
                args = [
                    _which('python3'),
                    '-m',
                    'mypy',
                    '--show-column-numbers',
                    '--strict',
                    '--no-incremental',
                    '--follow-imports=silent',
                    filename,
                ]
                try:
                    logging.debug('lint_python: Running %s', args)
                    subprocess.run(args,
                                   check=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   universal_newlines=True)
                except subprocess.CalledProcessError as cpe:
                    diagnostics.extend(
                        _process_diagnostics_output(filename,
                                                    lines,
                                                    cpe.output.strip(),
                                                    toolname='mypy'))

        if diagnostics:
            diagnostics.sort(key=lambda d: d.lineno or 0)
            raise LinterException('Python lint errors',
                                  fixable=False,
                                  diagnostics=diagnostics)
        return SingleResult(contents, [])

    @property
    def name(self) -> Text:
        return 'python'


class ClangFormatLinter(Linter):
    '''Runs clang-format.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        with tempfile.TemporaryDirectory(
                prefix='clang_format_linter_') as tmpdir:
            tmp_path = os.path.join(tmpdir, os.path.basename(filename))
            with open(tmp_path, 'wb') as outf:
                outf.write(contents)

            args = [
                _which('clang-format'),
                '--style=file',
                '--fallback-style=Google',
                f'--assume-filename={filename}',
                tmp_path,
            ]
            logging.debug('lint_clang_format: Running %s', args)
            result = subprocess.run(args,
                                    check=False,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
            if result.returncode != 0:
                raise LinterException('Found clang-format errors',
                                      diagnostics=[
                                          Diagnostic(
                                              result.stderr.decode('utf-8'),
                                              filename=filename),
                                      ],
                                      fixable=False)
            return SingleResult(result.stdout, ['clang-format'])

    @property
    def name(self) -> Text:
        return 'clang-format'


class EslintLinter(Linter):
    '''Runs eslint.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        del options
        self.__lock = threading.Lock()
        self.__eslint_state: Optional[Tuple[int, str]] = None

    def _get_port_and_token(self) -> Tuple[int, str]:
        if self.__eslint_state is None:
            with self.__lock:
                if self.__eslint_state is None:
                    args = [
                        _which('npx'),
                        'eslint_d',
                        'start',
                    ]
                    logging.debug('lint_eslint: Running %s', args)
                    subprocess.check_call(args,
                                          env={
                                              **os.environ, 'HOME': '/tmp'
                                          })
                    with open('/tmp/.eslint_d', encoding='utf-8') as portfile:
                        port_str, token = portfile.read().strip().split()
                    self.__eslint_state = (int(port_str), token)
        return self.__eslint_state

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        port, token = self._get_port_and_token()
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.connect(('localhost', port))
            args = [
                token,
                os.getcwd(),
                '--fix-dry-run',
                '--stdin',
                '--stdin-filename',
                filename,
                '--format=json',
                '--cache',
            ]
            logging.debug('lint_eslint: Running %s', args)
            sock.sendall(shlex.join(args).encode('utf-8') + b'\n')
            sock.sendall(contents)
            sock.shutdown(socket.SHUT_WR)

            stdout_chunks: List[bytes] = []
            while True:
                chunk = sock.recv(4096)
                if not chunk:
                    break
                stdout_chunks.append(chunk)
            stdout = (b''.join(stdout_chunks)).decode('utf-8')
            try:
                reports = json.loads(stdout)
            except json.decoder.JSONDecodeError as jde:
                # If there's extra data, let's just ignore it and try once
                # more.
                if jde.msg != 'Extra data':
                    raise
                reports = json.loads(stdout[:jde.pos])
            assert len(reports) == 1, reports
            report = reports[0]

            if len(report['messages']) > 0:
                lines = contents.decode('utf-8').split('\n')
                diagnostics = [
                    Diagnostic(
                        message=(f'[eslint] {message["message"]} '
                                 f'[{message["ruleId"]}]'),
                        filename=filename,
                        line=lines[message['line'] - 1],
                        lineno=message['line'],
                        col=message['column'],
                        col_end=(message['endColumn']
                                 if 'endColumn' in message else None),
                    ) for message in report['messages']
                ]
                raise LinterException('Eslint lint errors',
                                      fixable=False,
                                      diagnostics=diagnostics)
            if 'output' in report:
                # There were fixable errors.
                return SingleResult(report['output'].encode('utf-8'),
                                    ['eslint'])

            return SingleResult(contents, [])

    @property
    def name(self) -> Text:
        return 'eslint'


class CustomLinter(Linter):
    '''A lazily, dynamically-loaded linter.'''

    def __init__(self, custom_linter: Options,
                 config_file_path: Text) -> None:
        super().__init__()
        self.__module_path = os.path.join(
            os.path.dirname(config_file_path), custom_linter['path'])
        self.__config = custom_linter
        self.__instance: Optional[Linter] = None
        self.__options: Dict[Text, Text] = {}

    @property
    def _instance(self) -> Linter:
        if self.__instance is not None:
            return self.__instance
        custom_linter_module_spec = importlib.util.spec_from_file_location(
            self.__module_path.rstrip('.py').replace('/', '_'),
            self.__module_path)
        if custom_linter_module_spec is None:
            raise RuntimeError(f'Cannot find module {self.__module_path!r}')
        custom_linter_module = importlib.util.module_from_spec(
            custom_linter_module_spec)
        custom_linter_module_spec.loader.exec_module(  # type: ignore
            custom_linter_module)
        self.__instance = getattr(custom_linter_module,
                                  self.__config['class_name'])(self.__options)
        return self.__instance

    def __call__(self, options: Optional[Options] = None
                 ) -> 'CustomLinter':
        # Instead of the constructor being stored in the map of available
        # linters, a live instance of this class is stored. Later, this
        # function is called in lieu of the constructor.
        if options:
            self.__options = dict(options)
        return self

    def __getstate__(self) -> Mapping[Text, Any]:
        # This is needed to prevent self.__instance from being pickled, which
        # is something that is not supported.
        return {
            '_CustomLinter__config': self.__config,
            '_CustomLinter__module_path': self.__module_path,
            '_CustomLinter__options': self.__options,
            '_CustomLinter__instance': None,
        }

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        return self._instance.run_one(filename, contents)

    def run_all(self, filenames: Sequence[Text],
                contents_callback: Callable[[Text], bytes]) -> MultipleResults:
        return self._instance.run_all(filenames, contents_callback)

    @property
    def name(self) -> Text:
        return self._instance.name


class CommandLinter(Linter):
    '''Runs a custom command as linter.'''

    # pylint: disable=R0903

    def __init__(self, options: Optional[Options] = None) -> None:
        super().__init__()
        self.__options = options or {}

    def run_one(self, filename: str, contents: bytes) -> SingleResult:
        extension = os.path.splitext(filename)[1]
        with tempfile.NamedTemporaryFile(suffix=extension) as tmp:
            tmp.write(contents)
            tmp.flush()

            commands = [
                _custom_command(command, tmp.name, filename)
                for command in self.__options.get('commands', [])
            ]

            for args in commands:
                try:
                    logging.debug('lint_command: Running %s', args)
                    subprocess.run(args,
                                   check=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   universal_newlines=True)
                except subprocess.CalledProcessError as cpe:
                    raise LinterException(cpe.output, fixable=False) from cpe

            with open(tmp.name, 'rb') as tmp_in:
                return SingleResult(tmp_in.read(), ['command'])

    @property
    def name(self) -> Text:
        return f'command ({self.__options.get("commands", [])})'


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python3

"""Trivial linter that converts to uppercase."""


import sys


def main() -> None:
    """Main entrypoint."""

    filename = sys.argv[1]

    with open(filename, 'r', encoding='utf-8') as input_file:
        contents = input_file.read()

    with open(filename, 'w', encoding='utf-8') as output_file:
        output_file.write(contents.upper())


if __name__ == '__main__':
    main()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Main entrypoint for the omegaUp linting tool."""

from __future__ import print_function

import argparse
import concurrent.futures
import enum
import json
import logging
import os.path
import re
import sys
import traceback
from typing import (Any, Callable, Dict, Iterator, List, Mapping, Optional,
                    Sequence, Set, Text, Tuple)

from . import linters, git_tools

LinterFactory = Callable[..., linters.Linter]

_LINTER_MAPPING: Dict[Text, LinterFactory] = {
    'clang-format': linters.ClangFormatLinter,
    'eslint': linters.EslintLinter,
    'html': linters.HTMLLinter,
    'javascript': linters.JavaScriptLinter,
    'json': linters.JsonLinter,
    'karel': linters.KarelLinter,
    'markdown': linters.MarkdownLinter,
    'php': linters.PHPLinter,
    'problematic-terms': linters.ProblematicTermsLinter,
    'python': linters.PythonLinter,
    'style': linters.StyleLinter,
    'typescript': linters.TypeScriptLinter,
    'vue': linters.VueLinter,
    'whitespace': linters.WhitespaceLinter,
}

_ROOT = git_tools.root_dir()


class DiagnosticsOutput(enum.Enum):
    '''Controls the format in which diagnostic messages are displayed.'''
    STDERR = 'stderr'
    GITHUB = 'github'

    @staticmethod
    def values() -> Sequence['DiagnosticsOutput']:
        '''Returns the possible values of the enum.'''
        return (DiagnosticsOutput.STDERR, DiagnosticsOutput.GITHUB)


def _run_linter_one(
        linter: linters.Linter, filename: Text, contents: bytes,
        validate_only: bool,
        diagnostics_output: DiagnosticsOutput) -> Tuple[Optional[Text], bool]:
    '''Runs the linter against one file.'''
    try:
        try:
            new_contents, violations = linter.run_one(filename, contents)
        except linters.LinterException:
            raise
        except:  # noqa: bare-except
            exc_type, exc_value, exc_traceback = sys.exc_info()
            raise linters.LinterException(
                'Linter error',
                fixable=False,
                diagnostics=[
                    linters.Diagnostic(
                        message=''.join(
                            traceback.format_exception(exc_type, exc_value,
                                                       exc_traceback)),
                        filename=filename,
                    ),
                ],
            ) from None
    except linters.LinterException as lex:
        _report_linter_exception(filename, lex, diagnostics_output)
        return filename, lex.fixable

    if contents == new_contents:
        return None, False

    return _report_linter_results(filename, new_contents, validate_only,
                                  violations, True)


def _run_linter_all(
        args: argparse.Namespace, linter: linters.Linter,
        files: Sequence[Text], validate_only: bool,
        diagnostics_output: DiagnosticsOutput
) -> Sequence[Tuple[Optional[Text], bool]]:
    try:
        try:
            new_file_contents, original_contents, violations = linter.run_all(
                files, lambda filename: git_tools.file_contents(
                    args, _ROOT, filename))
        except linters.LinterException:
            raise
        except:  # noqa: bare-except
            exc_type, exc_value, exc_traceback = sys.exc_info()
            raise linters.LinterException(
                'Linter error',
                fixable=False,
                diagnostics=[
                    linters.Diagnostic(
                        message=''.join(
                            traceback.format_exception(exc_type, exc_value,
                                                       exc_traceback)),
                        filename=', '.join(files),
                    ),
                ],
            ) from None
    except linters.LinterException as lex:
        _report_linter_exception(', '.join(files), lex, diagnostics_output)
        return [(filename, lex.fixable) for filename in files]

    result: List[Tuple[Optional[Text], bool]] = []
    for filename in new_file_contents:
        if original_contents[filename] == new_file_contents[filename]:
            result.append((None, False))
        else:
            result.append(_report_linter_results(filename,
                                                 new_file_contents[filename],
                                                 validate_only, violations,
                                                 True))
    return result


def _report_linter_results(filename: Text, new_contents: bytes, validate: bool,
                           violations: Sequence[Text],
                           fixable: bool) -> Tuple[Text, bool]:
    violations_message = ', '.join(
        f'{git_tools.COLORS.FAIL}{violation}{git_tools.COLORS.NORMAL}'
        for violation in violations)
    if validate:
        print(('File '
               f'{git_tools.COLORS.HEADER}{filename}{git_tools.COLORS.NORMAL} '
               f'lint failed: {violations_message}'),
              file=sys.stderr)
    else:
        print(('Fixing '
               f'{git_tools.COLORS.HEADER}{filename}{git_tools.COLORS.NORMAL} '
               f'({", ".join(violations)})'),
              file=sys.stderr)
        with open(os.path.join(_ROOT, filename), 'wb') as outfile:
            outfile.write(new_contents)
    return filename, fixable


def _report_linter_exception(filename: Text, lex: linters.LinterException,
                             diagnostics_output: DiagnosticsOutput) -> None:
    if diagnostics_output == DiagnosticsOutput.GITHUB:
        for diagnostic in lex.diagnostics:
            location = [f'file={diagnostic.filename}']
            if diagnostic.lineno is not None:
                location.append(f'line={diagnostic.lineno}')
            if diagnostic.col is not None:
                location.append(f'col={diagnostic.col}')
            message = diagnostic.message.replace('%', '%25').replace(
                '\r', '%0D').replace('\n', '%0A')
            print((f'::{diagnostic.level} '
                   f'{",".join(location)}::{message}\n'),
                  end='')
    message_lines: List[str] = []
    if not lex.diagnostics:
        message_lines.append(
            'File '
            f'{git_tools.COLORS.FAIL}{filename}{git_tools.COLORS.NORMAL} '
            'lint failed:')
        message_lines.append(lex.message)
    for diagnostic in lex.diagnostics:
        location = [f'{git_tools.COLORS.FAIL}{diagnostic.filename}'
                    f'{git_tools.COLORS.NORMAL}']
        if diagnostic.lineno is not None:
            location.append(str(diagnostic.lineno))
            if diagnostic.col is not None:
                location.append(str(diagnostic.col))
        message_lines.append(f'{":".join(location)}: {diagnostic.message}')
        if diagnostic.line is not None:
            message_lines.append(f'    {diagnostic.line}')
            if diagnostic.col is not None:
                prefix = re.sub(r'[^\t]', ' ',
                                diagnostic.line[:diagnostic.col - 1])
                if diagnostic.col_end is not None:
                    arrow = '^' * (diagnostic.col_end - diagnostic.col)
                else:
                    arrow = '^'
                message_lines.append(f'    {prefix}{arrow}')
    message_lines.append('')
    print('\n'.join(message_lines), end='', file=sys.stderr)


def _report_error(message: Text,
                  diagnostics_output: DiagnosticsOutput) -> None:
    '''Display an error message to the user.'''
    if diagnostics_output == DiagnosticsOutput.GITHUB:
        message = message.replace('%',
                                  '%25').replace('\r',
                                                 '%0D').replace('\n', '%0A')
        print(f'::error ::{message}\n', end='')
    else:
        print(f'{git_tools.COLORS.FAIL}{message}{git_tools.COLORS.NORMAL}',
              file=sys.stderr)


def _run_linter(
        args: argparse.Namespace, linter: linters.Linter,
        filenames: Sequence[Text], validate_only: bool,
        diagnostics_output: DiagnosticsOutput) -> Tuple[Set[Text], bool]:
    '''Runs the linter against all files.'''
    logging.debug('%s: Files to consider: %s', linter.name,
                  ' '.join(filenames))
    logging.debug('%s: Running with %d threads', linter.name, args.jobs)
    files = dict((filename, git_tools.file_contents(args, _ROOT, filename))
                 for filename in filenames)
    with concurrent.futures.ThreadPoolExecutor(
            max_workers=args.jobs) as executor:
        futures = [
            executor.submit(_run_linter_one, linter, filename, contents,
                            validate_only, diagnostics_output)
            for filename, contents in files.items()
        ]
        results = [
            f.result() for f in concurrent.futures.as_completed(futures)
        ]
    results.extend(
        _run_linter_all(args, linter, filenames, validate_only,
                        diagnostics_output))
    return (set(violation for violation, _ in results
                if violation is not None),
            any(fixable for _, fixable in results))


def _get_enabled_linters(
        config: Mapping[Text, Any], config_file_path: Text,
        linter_allowlist: Text
) -> Iterator[Tuple[LinterFactory, Mapping[Text, Any]]]:
    '''Loads any custom linters.'''
    available_linters = dict(_LINTER_MAPPING)

    for custom_linter in config.get('custom_linters', []):
        available_linters[custom_linter['name']] = linters.CustomLinter(
            custom_linter, config_file_path)

    final_linter_allowlist = set(available_linters.keys())

    if linter_allowlist:
        args_linters = set(linter_allowlist.split(','))
        unknown_linters = args_linters - final_linter_allowlist
        if unknown_linters:
            print(('Unknown linters '
                   f'{git_tools.COLORS.FAIL}{", ".join(unknown_linters)}'
                   f'{git_tools.COLORS.NORMAL}.'),
                  file=sys.stderr)
            sys.exit(1)
        final_linter_allowlist = args_linters

    unknown_linters = set(config['lint']) - set(available_linters)
    if unknown_linters:
        print(('Unknown linters '
               f'{git_tools.COLORS.FAIL}{", ".join(unknown_linters)}'
               f'{git_tools.COLORS.NORMAL}.'),
              file=sys.stderr)
        sys.exit(1)

    for linter_name, options in config['lint'].items():
        if linter_name not in final_linter_allowlist:
            continue
        yield available_linters[linter_name], options


def main() -> None:
    '''Runs the linters against the chosen files.'''

    args = git_tools.parse_arguments(
        tool_description='lints a project',
        extra_arguments=[
            git_tools.Argument(
                '--pre-upload',
                action='store_true',
                help='Mark this as being run from within a pre-upload hook'),
            git_tools.Argument(
                '--command-name',
                default=None,
                type=str,
                help='Override the command name to execute this'),
            git_tools.Argument(
                '--linters', help='Comma-separated subset of linters to run'),
            git_tools.Argument(
                '--diagnostics-output',
                default=DiagnosticsOutput.STDERR,
                type=DiagnosticsOutput,
                choices=DiagnosticsOutput.values(),
                help='How to display diagnostics provided by the linters.'),
        ])
    if not args.files:
        return

    # If running in an automated environment, we can close stdin.
    # This will disable all prompts.
    if (args.continuous_integration
            or os.environ.get('CONTINUOUS_INTEGRATION') == 'true'):
        sys.stdin.close()

    validate_only = args.tool == 'validate'

    with open(args.config_file, 'r', encoding='utf-8') as config_file:
        config = json.load(config_file)

    file_violations: Set[Text] = set()
    fixable = False

    for linter, options in _get_enabled_linters(config, args.config_file,
                                                args.linters):
        filtered_files = args.files

        # Filter only the files in the allowlist.
        allowlist = [re.compile(r) for r in options.get('allowlist', [])]
        filtered_files = [
            filename for filename in filtered_files
            if any(r.match(filename) for r in allowlist)]

        # And not in the denylist.
        denylist = [re.compile(r) for r in options.get('denylist', [])]
        filtered_files = [
            filename for filename in filtered_files
            if all(not r.match(filename) for r in denylist)]
        local_violations, local_fixable = _run_linter(
            args, linter(options), filtered_files,
            validate_only, args.diagnostics_output)
        file_violations |= local_violations
        fixable |= local_fixable

    if file_violations:
        if not fixable:
            _report_error('Errors cannot be automatically fixed.',
                          args.diagnostics_output)
        elif validate_only:
            if git_tools.attempt_automatic_fixes(sys.argv[0],
                                                 args,
                                                 file_violations,
                                                 pre_upload=args.pre_upload):
                sys.exit(1)
            _report_error(
                ('Linter validation errors. '
                 'Please run\n\n    '
                 f'{git_tools.get_fix_commandline(args, file_violations)}\n\n'
                 'to fix them.'), args.diagnostics_output)
        else:
            print(('Files written to working directory. '
                   f'{git_tools.COLORS.HEADER}Please commit them '
                   f'before pushing.{git_tools.COLORS.NORMAL}'),
                  file=sys.stderr)
        sys.exit(1)

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Unit tests for git_tools."""

from __future__ import print_function

import unittest

from omegaup_hook_tools import git_tools


class TestGitTools(unittest.TestCase):
    """Tests git tools."""

    def test_get_explicit_file_list(self) -> None:
        """Tests git_tools.get_explicit_file_list()."""

        for commits, expected in [
                # Explicit separator.
                (['--', 'HEAD', 'foo'], ([], ['HEAD', 'foo'])),
                (['HEAD', '--', 'foo'], (['HEAD'], ['foo'])),
                (['HEAD', 'foo', '--'], (['HEAD', 'foo'], [])),

                # Sniffing.
                (['HEAD'], (['HEAD'], [])),
                (['foo'], ([], ['foo'])),
                (['HEAD', 'foo'], (['HEAD'], ['foo'])),
                (['HEAD', 'foo', 'HEAD'], (['HEAD'], ['foo', 'HEAD'])),
                (['HEAD', 'HEAD', 'foo'], (['HEAD', 'HEAD'], ['foo'])),
        ]:
            files = git_tools.get_explicit_file_list(commits)
            self.assertEqual((commits, files), expected)


if __name__ == '__main__':
    unittest.main()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python3

'''
Utility functions used to write git hooks.
'''

import argparse
import itertools
import logging
import multiprocessing
import os
import os.path
import re
import shlex
import subprocess
import sys
from typing import (Any, Iterable, Iterator, List, Mapping, Optional, Sequence,
                    Text)


HOOK_TOOLS_ROOT = os.path.abspath(os.path.join(__file__, '..'))
GIT_DIFF_TREE_PATTERN = re.compile(
    br'^:\d+ (\d+) [0-9a-f]+ [0-9a-f]+ ([ACDMRTUX])\d*$')
GIT_LS_TREE_PATTERN = re.compile(br'^\d* blob [0-9a-f]+\t(.*)$')
GIT_NULL_HASH = '0000000000000000000000000000000000000000'
GIT_DIRECTORY_ENTRY_MODE = b'160000'


class COLORS:
    '''Constants for colors in bash.'''
    # pylint: disable=R0903

    HEADER = '\033[95m'
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    NORMAL = '\033[0m'


class Argument:
    '''Class that represents a single argument for argparse.ArgumentParser.'''
    # pylint: disable=R0903

    def __init__(self, *args: Any, **kwargs: Any):
        self.args = args
        self.kwargs = kwargs

    def add_to(self, parser: argparse.ArgumentParser) -> None:
        '''Adds an additional argument to the parser.'''

        parser.add_argument(*self.args, **self.kwargs)


def get_explicit_file_list(commits: List[Text]) -> Sequence[Text]:
    '''Returns the explicit file list from the commandline.

    Developers might want to use an explicit file list in case there is a file
    with the same name as a commit. The way git disambiguates is that arguments
    that come before -- are references, and the ones that come after are files.
    We use the same convention.
    '''
    # If a -- was explicitly passed, honor it and don't try to guess what is
    # what.
    try:
        idx = commits.index('--')
        files = commits[idx + 1:]
        commits[idx:] = []
        return files
    except ValueError:
        pass

    # Otherwise let git-rev-parse let us know what are revisions and we treat
    # everything following the first non-revision as a file.
    try:
        commit_refs_output = subprocess.run(
            ['/usr/bin/git', 'rev-parse', '--revs-only'] + commits,
            universal_newlines=True,
            check=True,
            stdout=subprocess.PIPE,
            cwd=root_dir()).stdout.strip()
        if not commit_refs_output:
            commit_refs: List[Text] = []
        else:
            commit_refs = commit_refs_output.split('\n')
    except subprocess.CalledProcessError:
        commit_refs = []

    files = commits[len(commit_refs):]
    del commits[len(commit_refs):]
    return files


def _validate_args(args: argparse.Namespace, files: Sequence[Text]) -> bool:
    '''Validates whether args is valid.

    args.commits is valid if it has one commit (diffing from that commit
    against the working tree) or two commits.
    '''
    if args.all_files:
        if args.commits or files:
            print((f'{COLORS.FAIL}--all-files is incompatible '
                   f'with `commits` or `files`.{COLORS.NORMAL}'),
                  file=sys.stderr)
            return False
    if len(args.commits) not in (0, 1, 2):
        # args.commits can never be empty since its default value is ['HEAD'],
        # but the user can specify zero commits.
        print((f'{COLORS.FAIL}Can only specify zero, one or '
               f'two commits.{COLORS.NORMAL}'),
              file=sys.stderr)
        return False
    return True


def _get_all_files() -> Iterator[bytes]:
    '''Returns the list of all files at HEAD (and maybe in the index).'''

    for path in subprocess.run(['/usr/bin/git', 'ls-files', '-z'],
                               check=True,
                               stdout=subprocess.PIPE,
                               cwd=root_dir()).stdout.split(b'\x00'):
        if os.path.isfile(path):
            yield path


def _get_changed_files(commits: List[Text]) -> Iterator[bytes]:
    ''' Returns the list of files that were modified in the specified range.'''

    if not commits:
        cmd = ['/usr/bin/git', 'diff-index', '-z', '--diff-filter=d', 'HEAD']
    elif len(commits) == 1:
        cmd = ['/usr/bin/git', 'diff-index', '-z', '--diff-filter=d'] + commits
    else:
        if commits[-1] == GIT_NULL_HASH:
            # If the second commit is the null hash, the branch is being
            # deleted, so no files should be considered.
            return
        cmd = ['/usr/bin/git', 'diff-tree', '-r', '-z',
               '--diff-filter=d'] + commits
    tokens = subprocess.run(cmd,
                            check=True,
                            stdout=subprocess.PIPE,
                            cwd=root_dir()).stdout.split(b'\x00')
    idx = 0
    while idx < len(tokens) - 1:
        match = GIT_DIFF_TREE_PATTERN.match(tokens[idx])
        assert match, tokens[idx]
        filemode, status = match.groups()
        if filemode == GIT_DIRECTORY_ENTRY_MODE:
            # Files with the 160000 mode are not actually files or
            # directories.  They just are directory entries, and they
            # typically appear in the path where submodules are inserted
            # into the tree.
            idx += 2
            continue
        src = tokens[idx + 1]
        if status in (b'C', b'R'):
            dest = tokens[idx + 2]
            idx += 3
            yield dest
        else:
            idx += 2
            yield src


def _files_to_consider(args: argparse.Namespace) -> List[Text]:
    '''Returns the list of files to consider, based on |args|' commits.'''

    # Get all files in the latter commit.
    if args.all_files:
        result = _get_all_files()
    else:
        result = _get_changed_files(args.commits)

    return sorted([str(filename, encoding='utf-8') for filename in result])


def prompt(question: Text, default: bool = True) -> bool:
    '''Asks the user a yes/no question.'''
    if sys.stdin.closed or not sys.stdin.isatty():
        return default

    while True:
        yes_str = 'yes'
        no_str = 'no'
        yes_label = yes_str
        no_label = no_str
        if default:
            yes_label = yes_label.upper()
        else:
            no_label = no_label.upper()

        try:
            response = input(f'{question} ({yes_label}/{no_label}): ')
        except EOFError:
            return default

        response = response.strip().lower()
        if not response:
            break
        if yes_str.startswith(response):
            return True
        if no_str.startswith(response):
            return False

    return default


def file_contents(args: argparse.Namespace, root: Text,
                  filename: Text) -> bytes:
    '''Returns contents of |filename| at the revision specified by |args|.'''
    if len(args.commits) in (0, 1):
        # Zero or one commits (where the former is a shorthand for 'HEAD')
        # always diff against the current contents of the file in the
        # filesystem.
        with open(os.path.join(root, filename), 'rb') as working_dir_file:
            return working_dir_file.read()
    else:
        return subprocess.run(
            ['/usr/bin/git', 'show', f'{args.commits[-1]}:{filename}'],
            check=True,
            stdout=subprocess.PIPE,
            cwd=root).stdout


def root_dir() -> Text:
    '''Returns the top-level directory of the project.'''
    return subprocess.run(['/usr/bin/git', 'rev-parse', '--show-toplevel'],
                          universal_newlines=True,
                          check=True,
                          stdout=subprocess.PIPE).stdout.strip()


def parse_arguments(
        tool_description: Optional[Text] = None,
        extra_arguments: Sequence[Argument] = ()) -> argparse.Namespace:
    '''Parses the commandline arguments.'''
    parser = argparse.ArgumentParser(description=tool_description)
    parser.add_argument(
        '--verbose', '-v', action='store_true',
        help='Prints verbose information')
    parser.add_argument(
        '--config-file',
        default=os.path.join(root_dir(), '.lint.config.json'),
        help='Prints verbose information')
    parser.add_argument(
        '--continuous-integration', action='store_true',
        help=('Assumes this is an unsupervised environment. '
              'Disables all prompts.'))
    parser.add_argument(
        '--jobs', '-j', type=int, help='Number of parallel jobs',
        default=multiprocessing.cpu_count())
    for argument in extra_arguments:
        argument.add_to(parser)
    subparsers = parser.add_subparsers(dest='tool')
    subparsers.required = True

    validate_parser = subparsers.add_parser(
        'validate', help='Only validates, does not make changes')
    validate_parser.add_argument(
        '--all-files', action='store_true',
        help='Considers all files. Incompatible with `commits` and `files`')
    validate_parser.add_argument(
        'commits',
        metavar='[commit [commit ...]] [--] [file [file ...]]',
        nargs=argparse.REMAINDER,
        default=[],
        type=str,
        help=('commit: Only include files changed between commits\n'
              'file:   If specified, only consider these files'))

    fix_parser = subparsers.add_parser(
        'fix',
        help=('Fixes all violations and leaves '
              'the results in the working tree.'))
    fix_parser.add_argument(
        '--all-files', action='store_true',
        help=('Considers all files. '
              'Incompatible with `commits` and `files`'))
    fix_parser.add_argument(
        'commits',
        metavar='[commit [commit ...]] [--] [file [file ...]]',
        nargs=argparse.REMAINDER,
        default=[],
        type=str,
        help=('commit: Only include files changed between commits\n'
              'file:   If specified, only consider these files'))

    subparsers.add_parser(
        'ensure-container',
        help=('Does nothing. '
              'Allows the container to be downloaded in a separate command.'))

    args = parser.parse_args()
    if args.tool == 'ensure-container':
        sys.exit(0)

    files = get_explicit_file_list(args.commits)
    if not _validate_args(args, files):
        sys.exit(1)
    if files:
        args.files = files
    else:
        args.files = list(_files_to_consider(args))

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
    logging.debug('Files to consider: %s', ' '.join(args.files))
    return args


def _get_quoted_command_name(command_name: Optional[Text]) -> List[Text]:
    '''Returns the name of the command needed to invoke this script.'''
    if command_name is not None:
        return [shlex.quote(command_name)]
    if os.environ.get('DOCKER') == 'true':
        return [
            '/usr/bin/docker', 'run', '--rm', '-v', '"$PWD:/src"', '-v',
            '"$PWD:$PWD"', 'omegaup/hook_tools'
        ]
    return [shlex.quote(sys.argv[0])]


def _get_fix_args(prog_args: List[Text],
                  args: argparse.Namespace,
                  files: Optional[Iterable[Text]] = None) -> Sequence[Text]:
    '''Gets the command arguments to run to fix violations.'''
    params = prog_args + ['fix']
    params.extend(args.commits)
    if not files and args.files:
        files = args.files
    if files:
        params.append('--')
        params.extend(sorted(files))
    return params


def get_fix_commandline(args: argparse.Namespace,
                        files: Optional[Iterable[Text]] = None) -> Text:
    '''Gets the commandline the developer must run to fix violations.'''
    return ' '.join(
        itertools.chain(
            _get_quoted_command_name(args.command_name if 'command_name' in
                                     args else None),
            (shlex.quote(p) for p in _get_fix_args([], args, files)),
        ))


def verify_toolchain(binaries: Mapping[Text, Text]) -> bool:
    '''Verifies that the developer has all necessary tools installed.'''
    success = True
    for path, install_cmd in binaries.items():
        if not os.path.isfile(path):
            print((f'{COLORS.FAIL}{path} not found.{COLORS.NORMAL} '
                   f'Please run `{install_cmd}` to install.'),
                  file=sys.stderr)
            success = False
    return success


def _is_single_commit_pushed(args: argparse.Namespace) -> bool:
    '''Returns whether a single commit is being pushed.'''
    if not args.commits:
        # This case is where only HEAD is considered, but we don't know what
        # the merge base is.
        return False
    merge_base = str(args.commits[0])
    if len(args.commits) == 1:
        pushed_commit = 'HEAD'
    else:
        pushed_commit = args.commits[1]
    return merge_base == subprocess.run(
        [
            '/usr/bin/git',
            'rev-parse',
            f'{pushed_commit}^',
        ],
        universal_newlines=True,
        check=True,
        stdout=subprocess.PIPE).stdout.strip()


def attempt_automatic_fixes(scriptname: Text,
                            args: argparse.Namespace,
                            files: Optional[Iterable[Text]] = None,
                            pre_upload: bool = False) -> bool:
    '''Attempts to automatically fix any fixable errors.'''
    if sys.stdin.closed or not sys.stdin.isatty():
        # There is no one to ask.
        return False
    if not prompt('Want to automatically fix errors?'):
        # User decided not to go with the fixes.
        return False
    # This should always "fail" because it's designed to block `git push`.  We
    # cannot use check_call() for that reason. We also always use the
    # in-container version of the invocation.
    subprocess.call(_get_fix_args([scriptname], args, files))
    if not subprocess.run(['/usr/bin/git', 'status', '--porcelain'],
                          check=True,
                          stdout=subprocess.PIPE).stdout.strip():
        # The fix failed?
        return False
    if pre_upload:
        continue_message = 'please run the `git push` command again.'
    else:
        continue_message = 'ready to upload.'
    if not prompt('Want to also commit the fixes?'):
        # Fixes succeeded, even if they are not committed yet.
        print(
            ('Files written to working directory. '
             f'{COLORS.HEADER}Please commit them '
             f'before pushing.{COLORS.NORMAL}'),
            file=sys.stderr,
        )
        return True
    if _is_single_commit_pushed(args):
        # We can amend the previous commit!
        commit_params = ['/usr/bin/git', 'commit', '--amend', '--no-edit']
        if files:
            commit_params.append('--')
            commit_params.extend(files)
        else:
            commit_params.append('--all')
        subprocess.run(commit_params, check=True)
        print((f'{COLORS.OKGREEN}Previous commit reused, {continue_message}'
               f'{COLORS.NORMAL}'),
              file=sys.stderr)
    else:
        commit_params = ['/usr/bin/git', 'commit',
                         '-m', f'Fixed {scriptname} lints']
        if files:
            commit_params.append('--')
            commit_params.extend(files)
        else:
            commit_params.append('--all')
        subprocess.run(commit_params, check=True)
        print((f'{COLORS.OKGREEN}Committed fixes, {continue_message}'
               f'{COLORS.NORMAL}'),
              file=sys.stderr)
    return True


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Unit tests for the Linters."""

from __future__ import print_function

import unittest

from omegaup_hook_tools import linters


class TestLinters(unittest.TestCase):
    """Tests the linters."""

    def test_whitespace(self) -> None:
        """Tests WhitespaceLinter."""

        linter = linters.WhitespaceLinter()

        self.assertEqual(
            linter.run_one('test.txt', b'Hello\r\nWorld!\n'),
            linters.SingleResult(b'Hello\nWorld!\n', ['Windows-style EOF']))

        self.assertEqual(
            linter.run_one('test.txt', b'Hello\n\n\nWorld!\n'),
            linters.SingleResult(b'Hello\n\nWorld!\n',
                                 ['consecutive empty lines']))

        self.assertEqual(
            linter.run_one('test.txt', b'function() {\n\n}\n'),
            linters.SingleResult(b'function() {\n}\n',
                                 ['empty lines after an opening brace']))

        self.assertEqual(
            linter.run_one('test.txt', b'function() {\n//\n\n}\n'),
            linters.SingleResult(b'function() {\n//\n}\n',
                                 ['empty lines before a closing brace']))

        self.assertEqual(
            linter.run_one('test.txt', 'hello \u200b\n'.encode('utf-8')),
            linters.SingleResult(b'hello\n', [
                'trailing whitespace',
            ]))

        self.assertEqual(
            linter.run_one('test.txt', b'function() {\r\n\n\n// \n\n}\n'),
            linters.SingleResult(b'function() {\n//\n}\n', [
                'Windows-style EOF',
                'trailing whitespace',
                'consecutive empty lines',
                'empty lines after an opening brace',
                'empty lines before a closing brace',
            ]))

    def test_command(self) -> None:
        """Tests CommandLinter."""

        linter = linters.CommandLinter({
            'commands': ['python3 tests/data/uppercase_linter.py'],
        })

        self.assertEqual(linter.run_one('test.txt', b'Hello, World!\n'),
                         linters.SingleResult(b'HELLO, WORLD!\n', ['command']))

    def test_javascript(self) -> None:
        """Tests JavaScriptLinter."""

        linter = linters.JavaScriptLinter()

        self.assertEqual(
            linter.run_one('test.js', b'  function x ( ){a;b;c;};\n'),
            linters.SingleResult(b'function x() {\n  a;\n  b;\n  c;\n}\n',
                                 ['javascript']))

        self.assertEqual(
            linter.run_one('test.js', b'#!/usr/bin/node\nreturn;\n'),
            linters.SingleResult(b'#!/usr/bin/node\nreturn;\n',
                                 ['javascript']))

    def test_vue(self) -> None:
        """Tests VueLinter."""

        linter = linters.VueLinter()

        with self.assertRaisesRegex(linters.LinterException,
                                    r'Found Vue errors') as lex:
            linter.run_one('test.vue',
                           b'<template>\n<b></span>\n</template>\n')
        self.assertEqual(lex.exception.diagnostics, [
            linters.Diagnostic(message='Unclosed tag',
                               filename='test.vue',
                               line='<b></span>',
                               lineno=2,
                               col=4),
            linters.Diagnostic('Unclosed tag',
                               filename='test.vue',
                               line='</template>',
                               lineno=3,
                               col=1),
        ])

        self.assertEqual(
            linter.run_one('test.vue', b'<template>\n<b></b>\n</template>\n'),
            linters.SingleResult(b'<template>\n  <b></b>\n</template>\n',
                                 ['vue']))

    def test_html(self) -> None:
        """Tests HTMLLinter."""

        linter = linters.HTMLLinter()

        self.assertEqual(
            linter.run_one(
                'test.html', b'<!DOCTYPE html>\n<html><head><title /></head>'
                b'<body>\n<input/></body></html>\n'),
            linters.SingleResult(
                b'<!DOCTYPE html>\n<html>\n  <head>\n    <title></title>\n'
                b'  </head>\n  <body>\n    <input />\n  </body>\n</html>\n',
                ['html']))

    def test_php(self) -> None:
        """Tests PHPLinter."""

        linter = linters.PHPLinter()

        self.assertEqual(
            linter.run_one('test.php', b'<?php\necho array("foo");'),
            linters.SingleResult(b'<?php\necho [\'foo\'];\n', ['php']))

    def test_python(self) -> None:
        """Tests PythonLinter."""

        linter = linters.PythonLinter()

        with self.assertRaisesRegex(linters.LinterException,
                                    'Python lint errors') as lex:
            linter.run_one('test.py', b'def main():\n  pass\n')
        self.assertEqual(lex.exception.diagnostics, [
            linters.Diagnostic(
                message=('[pylint] C0114(missing-module-docstring) '
                         'Missing module docstring'),
                filename='test.py',
                line='def main():',
                lineno=1,
            ),
            linters.Diagnostic(
                message=('[pylint] C0116(missing-function-docstring) '
                         'Missing function or method docstring'),
                filename='test.py',
                line='def main():',
                lineno=1,
            ),
            linters.Diagnostic(
                message=('[pycodestyle] E111 indentation is not a '
                         'multiple of 4'),
                filename='test.py',
                line='  pass',
                lineno=2,
                col=3,
            ),
            linters.Diagnostic(
                message=('[pylint] W0311(bad-indentation) '
                         'Bad indentation. Found 2 spaces, expected 4'),
                filename='test.py',
                line='  pass',
                lineno=2,
            ),
        ])


if __name__ == '__main__':
    unittest.main()

# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
